@@ -21,28 +21,91 @@ try:
 except:
   import pickle
 
+class MacroParser:
+    name = Word(alphas)
+    opts_delimiters = ('[',']')
+    options = originalTextFor( nestedExpr( *opts_delimiters ) )
+    args_delimiters = ('{','}')
+    arguments = originalTextFor( nestedExpr( *args_delimiters ) )
 
-class MacroProcessor(object):
+    macro = Combine( WordStart('\\') + Literal('\\') + name('name') + Optional(options('options')) + OneOrMore(arguments)('arguments') )
+    macro.parseWithTabs()
+
+def expand_macro( macro_processor, parsed_tokens ):
+    '''This function is called to expand a macro that has been parsed by pyparsing.
+       It recieves a MacroExpansionAction that contains everything need to perform the expansion.
+
+       It will recieve the matched elements as a ParseResult in the parsed_tokens argument.
+       The macro name of the macro that was found will be in parsed_tokens["name"]. The macro options and arguments will
+       be in parsed_tokens["options"] and parsed_tokens["arguments"]. The original text that was matched will be in parsed_tokens[0]
+       
+       It should process the macro and return the text that will replace it, or None.
+
+       If self.use_cache is True, the expanded text is written to cache. If the same macro is found again (it must be identical),
+       the cached value will be used for expansion.
+       '''
+
+    if macro_processor.use_cache and parsed_tokens[0] in macro_processor.cache:
+      return macro_processor.cache[parsed_tokens[0]]
+
+    name = str(parsed_tokens.name)
+    # options and arguments are nested expressions. the token we get
+    # will be wrapped in [] (for options) and {} (for arguments), so we need
+    # to strip them off.
+    options   = parsed_tokens.options[len(MacroParser.opts_delimiters[0]):-len(MacroParser.opts_delimiters[1])]
+    arguments = [   argument[len(MacroParser.args_delimiters[0]):-len(MacroParser.args_delimiters[1])] for argument in parsed_tokens.arguments]
+
+
+    # Order matters here. The first handler found will be used. Handlers are
+    # looked for in:
+    # 1. the user macros.py file.
+    # 2. the added_macros member (macro handlers added with the addMacro method)
+    # 3. our macros.py file
+    handler = macro_processor.macroCallbacks.get(name, None)
+
+    if handler is None:
+      return parsed_tokens[0]
+
+    nargs = macro_processor.macroCallbacksArity[name]
+    if nargs == 3:
+      replacement = handler(self,arguments,options)
+    elif nargs == 2:
+      replacement = handler(arguments,options)
+    elif nargs == 1:
+      replacement = handler(arguments)
+
+    if replacement is None:
+      return parsed_tokens[0]
+
+    if self.use_cache:
+      self.cache[parsed_tokens[0]] = replacement
+    
+
+    return replacement
+
+
+class MacroProcessor:
   def __init__(self,use_cache=False):
     # define macro grammer
-    name = Word(alphas)
-    self.opts_delimiters = ('[',']')
-    options = originalTextFor( nestedExpr( *self.opts_delimiters ) )
-    self.args_delimiters = ('{','}')
-    arguments = originalTextFor( nestedExpr( *self.args_delimiters ) )
-
-    self.macroParser = Combine( WordStart('\\') + Literal('\\') + name('name') + Optional(options('options')) + OneOrMore(arguments)('arguments') )
+    self.macroParser = MacroParser.macro
 
     self.added_macros = {}
 
     self.cache = {}
     self.use_cache = use_cache
 
+    self.macroCallbacks = {}
+    self.macroCallbacksArity = {}
+    self._updateMacroCallbacksWithOurMacros()
+    self._updateMacroCallbacksWithUserMacros()
+
   def addParserAction(self,func):
     self.macroParser.addParserAction(func)
 
   def addMacro(self,name,func):
     self.added_macros[name] = func
+    self.macroCallbacks[name] = func
+    self._updateMacroCallbacksWithUserMacros()
 
   def clearCache(self):
     '''Clear current cache.'''
@@ -62,6 +125,20 @@ class MacroProcessor(object):
     with open(fn,'rb') as fh:
       self.cache = pickle.load(fh)
 
+  def _updateMacroCallbacksWithOurMacros(self):
+    self.macroCallbacks = dict( inspect.getmembers(our_macros, lambda m: inspect.isfunction(m) and m.__module__ == our_macros.__name__ and not m.__name__.startswith("_")) )
+    self._updateMacroCallbacksArity()
+    
+
+  def _updateMacroCallbacksWithUserMacros(self):
+    if 'user_macros' in sys.modules:
+      self.macroCallbacks.update( dict( inspect.getmembers(user_macros, lambda m: inspect.isfunction(m) and m.__module__ == user_macros.__name__ and not m.__name__.startswith("_")) ) )
+      self._updateMacroCallbacksArity()
+
+  def _updateMacroCallbacksArity(self):
+    for k,v in self.macroCallbacks.items():
+      self.macroCallbacksArity[k] = len(inspect.signature(v).parameters)
+
 
   def process(self,text,repeat=True):
     '''Process a string. Macros found in the string that have handlers will be replaced with the output of the handler.'''
@@ -76,16 +153,33 @@ class MacroProcessor(object):
       # benchmarks indicate that this is actually fater than using pyparsing's transformString
       # and it allows for macro expansion to be done in parallel.
       results = self.macroParser.scanString( text )
-      parts = []
+      output_parts = []
+      parsed_macros = []
+      macro_part_mapping = []
       last_i = 0
-      for r in results:
-        parts.append(text[last_i:r[1]])
-        parts.append( (text[r[1]:r[2]],r[0])  )
-        last_i = r[2]
-      parts.append(text[last_i:])
+      for parsed_tokens,si,ei in results:
+        output_parts.append(text[last_i:si])
+        # only add macros we don't have a chached value for.
+        if self.use_cache and parsed_tokens[0] in self.cache:
+          output_parts.append(self.cache[parsed_tokens[0]])
+        else:
+          parsed_macros.append(parsed_tokens)
+          macro_part_mapping.append(len(output_parts))
+          output_parts.append(None)
+        last_i = ei
+      output_parts.append(text[last_i:])
 
-      # now process all of the parts, performing macro expansion on the macros that have been parsed.
-      newtext = "".join([self._process_part(p) for p in parts])
+      # process all of the macros
+      expanded_macros = map(self._expand_macro, parsed_macros)
+
+      for i,e in enumerate(expanded_macros):
+        output_parts[macro_part_mapping[i]] = e
+        if self.use_cache:
+          self.cache[parsed_macros[i][0]] = e
+    
+
+
+      newtext = "".join(output_parts)
 
 
       if newtext == text or repeat == False:
@@ -95,21 +189,11 @@ class MacroProcessor(object):
 
     return text
 
-  def _process_part(self,part):
-    if isinstance(part,tuple):
-      replace = self.expand(part[0],0,part[1])
-      if replace is None:
-        return part[0]
-      return replace
-
-    return part
-
-  def expand(self,match,loc,toks):
+  def _expand_macro(self,parsed_toks):
     '''This function is called to expand a macro that has been parsed by pyparsing.
-       It will recieve the matched elements as a dict in the toks argument. The original string, and
-       its location in the text are also padded (i.e., this function can be used as a parseAction). The macro name
-       of the macro that was found will be in toks["name"]. The macro options and arguments will
-       be in toks["options"] and toks["arguments"].
+       It will recieve the matched elements as a dict in the parsed_toks argument.
+       The macro name of the macro that was found will be in parsed_toks["name"]. The macro options and arguments will
+       be in parsed_toks["options"] and parsed_toks["arguments"]. The original text that was matched will be in parsed_toks[0]
        
        It should process the macro and return the text that will replace it, or None.
 
@@ -117,15 +201,15 @@ class MacroProcessor(object):
        the cached value will be used for expansion.
        '''
 
-    if self.use_cache and match in self.cache:
-      return self.cache[match]
+    if self.use_cache and parsed_toks[0] in self.cache:
+      return self.cache[parsed_toks[0]]
 
-    name = str(toks.name)
+    name = str(parsed_toks.name)
     # options and arguments are nested expressions. the token we get
     # will be wrapped in [] (for options) and {} (for arguments), so we need
     # to strip them off.
-    options   = toks.options[len(self.opts_delimiters[0]):-len(self.opts_delimiters[1])]
-    arguments = [   argument[len(self.args_delimiters[0]):-len(self.args_delimiters[1])] for argument in toks.arguments]
+    options   = parsed_toks.options[len(MacroParser.opts_delimiters[0]):-len(MacroParser.opts_delimiters[1])]
+    arguments = [   argument[len(MacroParser.args_delimiters[0]):-len(MacroParser.args_delimiters[1])] for argument in parsed_toks.arguments]
 
 
     # Order matters here. The first handler found will be used. Handlers are
@@ -133,15 +217,12 @@ class MacroProcessor(object):
     # 1. the user macros.py file.
     # 2. the added_macros member (macro handlers added with the addMacro method)
     # 3. our macros.py file
-    handler = lambda x,y,z : None
-    if 'user_macros' in sys.modules and hasattr(user_macros,name):
-      handler = getattr(user_macros,name)
-    elif name in self.added_macros:
-      handler = self.added_macros[name]
-    elif hasattr(our_macros,name):
-      handler = getattr(our_macros,name)
+    handler = self.macroCallbacks.get(name, None)
 
-    nargs=len(inspect.signature( handler ).parameters)
+    if handler is None:
+      return parsed_toks[0]
+
+    nargs = self.macroCallbacksArity[name]
     if nargs == 3:
       replacement = handler(self,arguments,options)
     elif nargs == 2:
@@ -149,9 +230,8 @@ class MacroProcessor(object):
     elif nargs == 1:
       replacement = handler(arguments)
 
-    if self.use_cache:
-      self.cache[match] = replacement
-
+    if replacement is None:
+      return parsed_toks[0]
 
     return replacement
 
